"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const removeOpSize = JSON.stringify({
    op: "remove",
    path: ""
}).length;
const replaceOpSize = JSON.stringify({
    op: "replace",
    path: "",
    value: ""
}).length - 2;
function isEmpty(data) {
    return _.isString(data) || _.isEmpty(data);
}
function getPatchPath(path) {
    if (path.length === 0) {
        return "";
    }
    else {
        let res = "";
        path.forEach((value) => {
            res += "/" + value;
        });
        return res;
    }
}
function pathSize(path) {
    return getPatchPath(path).length;
}
function removeSize(path) {
    return removeOpSize + pathSize(path) + 1;
}
function replaceSize(path, objSize) {
    return replaceOpSize + pathSize(path) + objSize + 1;
}
function addSize(path, objSize) {
    return replaceSize(path, objSize) - 4;
}
function getType(value) {
    if (_.isArray(value)) {
        return "ARRAY";
    }
    else if (_.isObject(value)) {
        return "OBJECT";
    }
    else {
        return "VALUE";
    }
}
function getSize(obj) {
    return JSON.stringify(obj).length;
}
function makeArrayTree(oldObj, newObj, path) {
    let op;
    const sizeDiff = newObj.length - oldObj.length;
    if (sizeDiff === 1) {
        op = "add";
    }
    else if (sizeDiff === -1) {
        op = "remove";
    }
    else {
        return false;
    }
    const maxLength = Math.max(oldObj.length, newObj.length);
    let children = [];
    let diffSize = 0;
    let oldIndex = 0;
    let newIndex = 0;
    let opAdded = false;
    for (let i = 0; i < maxLength; i += 1) {
        if (_.isEqual(newObj[newIndex], oldObj[oldIndex])) {
            oldIndex += 1;
            newIndex += 1;
        }
        else if (opAdded) {
            return false;
        }
        else {
            const newPath = path.concat(i);
            if (op === "add") {
                const childSize = getSize(newObj[newIndex]);
                const childDiffSize = addSize(newPath, childSize);
                diffSize += childDiffSize;
                children.push({
                    path: newPath,
                    size: childSize,
                    diffSize: childDiffSize,
                    operation: "add"
                });
                newIndex += 1;
            }
            else {
                const childDiffSize = removeSize(newPath);
                diffSize += childDiffSize;
                children.push({
                    path: newPath,
                    size: 0,
                    diffSize: childDiffSize,
                    operation: "remove"
                });
                oldIndex += 1;
            }
            opAdded = true;
        }
    }
    return {
        path,
        size: getSize(newObj),
        diffSize,
        children,
        operation: "children"
    };
}
function makeTree(oldObj, newObj, path = []) {
    if (newObj === undefined && oldObj !== undefined) {
        return {
            path,
            size: 0,
            diffSize: removeSize(path),
            operation: "remove"
        };
    }
    if (oldObj === undefined) {
        const size = getSize(newObj);
        return {
            path,
            size,
            diffSize: addSize(path, size),
            operation: "add"
        };
    }
    if (getType(newObj) !== getType(oldObj)) {
        const size = getSize(newObj);
        return {
            path,
            size,
            diffSize: replaceSize(path, size),
            operation: "replace"
        };
    }
    if (isEmpty(newObj) || (isEmpty(oldObj))) {
        const size = getSize(newObj);
        if (_.isEqual(newObj, oldObj)) {
            return {
                path,
                size,
                diffSize: 0,
                operation: "none"
            };
        }
        else {
            return {
                path,
                size,
                diffSize: replaceSize(path, size),
                operation: "replace"
            };
        }
    }
    if (_.isArray(newObj)) {
        const result = makeArrayTree(oldObj, newObj, path);
        if (result) {
            return result;
        }
    }
    const keys = _.uniq(_.keys(newObj).concat(_.keys(oldObj)));
    let size = 0;
    let children = [];
    let diffSize = 0;
    for (const key of keys) {
        const child = makeTree(oldObj[key], newObj[key], path.concat(key));
        size += child.size + key.length + 2;
        diffSize += child.diffSize;
        if (child.operation !== "none") {
            children.push(child);
        }
    }
    ;
    const replaceDiffSize = replaceSize(path, size);
    if (diffSize > replaceDiffSize) {
        return {
            path,
            size,
            diffSize: replaceDiffSize,
            operation: "replace"
        };
    }
    else {
        return {
            path,
            size,
            diffSize,
            children,
            operation: "children"
        };
    }
}
function get(newObj, path) {
    if (path.length === 0) {
        return newObj;
    }
    else {
        return _.get(newObj, path);
    }
}
function makeDiff(result, newObj, node) {
    switch (node.operation) {
        case "add":
            result.push({
                op: "add",
                path: getPatchPath(node.path),
                value: get(newObj, node.path)
            });
            break;
        case "remove":
            result.unshift({
                op: "remove",
                path: getPatchPath(node.path)
            });
            break;
        case "replace":
            result.push({
                op: "replace",
                path: getPatchPath(node.path),
                value: get(newObj, node.path)
            });
            break;
        case "children":
            node.children.forEach((child) => {
                makeDiff(result, newObj, child);
            });
            break;
    }
}
function diff(oldObj, newObj) {
    oldObj = JSON.parse(JSON.stringify(oldObj));
    newObj = JSON.parse(JSON.stringify(newObj));
    const result = [];
    const node = makeTree(oldObj, newObj);
    makeDiff(result, newObj, node);
    return result;
}
exports.diff = diff;
//# sourceMappingURL=index.js.map